import casadi as ca
import matplotlib.pyplot as plt
import numpy as np

from rk4step import rk4step


# for people who cannot see an interactive plot, uncomment the following lines
import matplotlib
if matplotlib.get_backend() == 'agg':
    matplotlib.use('WebAgg')
print(f'backend: {matplotlib.get_backend()}')


## parameters required for the OCP
nx = 2  # state dimension
nu = 1  # control dimension
N = 50  # horizon
dt = 0.1  # discretization time step
N_rk4 = 10  # number of rk4 steps per discretization time step
x0_bar = np.array([np.pi, 0])   # initial state

## build the integrator
# ode
def dynamics(x, u):
    return ca.vertcat(x[1], ca.sin(x[0]) + u)

# discrete dynamics
x = ca.SX.sym('x', nx, 1)
u = ca.SX.sym('u', nu, 1)
x_next = x
for _ in range(N_rk4):
    x_next = rk4step(x, u, dynamics, dt)

# integrator
F = ca.Function('F', [x, u], [x_next])


## NLP formulation
x_traj = ca.SX.sym('x_traj', nx*N, 1)  # vector of all states (1,..,N)
u_traj = ca.SX.sym('u_traj', nu*N, 1)  # vector of all controls (0,..,N-1)

x_traj_init = .1 * np.ones((nx*N, 1))  # intial guess
u_traj_init = .1 * np.ones((nu*N, 1))  # intial guess


# some utils
u_traj_per_stage = ca.vertsplit(u_traj, nu)
x_traj_per_stage = ca.vertsplit(x_traj, nx)
stage_cost = ca.Function('stage_cost', [x, u], [x.T @ x + 2 * u.T @ u])
terminal_cost = ca.Function('terminal_cost', [x], [10 * x.T @ x])
roh = .01
barrier_term = ca.Function('barrier', [u], [- roh * (ca.log(-u+1) + ca.log(u+1))])

# build cost and constraints
x_current = x0_bar
constraints = []
cost = 0
for i in range(N):
    u_current = u_traj_per_stage[i]
    x_next = x_traj_per_stage[i]

    constraints.append(x_next - F(x_current, u_current))
    cost += stage_cost(x_current, u_current)
    cost += barrier_term(u_current)

    x_current = x_next
# terminal cost
cost += terminal_cost(x_current)
constraints = ca.vertcat(*constraints)


## lagrangian function
lam_traj = ca.SX.sym('lam_traj', nx*N, 1)  # multiplier (1,..,N)
lagrangian = cost + lam_traj.T @ constraints

# plot sparsity pattern of the Hessian of the lagrangian
lam_traj_per_stage = ca.vertsplit(lam_traj, nx)
z_traj = []
for i in range(N):
    # primal and dual variables in the required order 
    u_current = u_traj_per_stage[i]
    lam = lam_traj_per_stage[i]
    x_next = x_traj_per_stage[i]
    z_traj.extend([u_current, lam, x_next])
z_traj = ca.vertcat(*z_traj)
(hess, grad) = ca.hessian(lagrangian, z_traj)
hess_lagrangian = ca.Function('hess_lagrangian', [z_traj], [hess])

plt.figure(1)
z_traj_mock = np.ones(z_traj.shape)
plt.spy(hess_lagrangian(z_traj_mock).full(), markersize=3)


## nlp solver
nlp_sim = {'x': ca.vertcat(x_traj, u_traj), 'f': cost, 'g':constraints}
solver_sim = ca.nlpsol('solver','ipopt', nlp_sim)

# solve nlp
sol_sim = solver_sim(
    x0=ca.vertcat(x_traj_init, u_traj_init),
    lbg=np.zeros(constraints.shape),
    ubg=np.zeros(constraints.shape),
)

## visualize solution
x_u_traj_opt_sim = sol_sim['x'].full().T
x_traj_opt_sim = ca.reshape(
    ca.vertcat(x0_bar, x_u_traj_opt_sim[:, 0:N*nx].T),
    (nx, N+1)
)
u_traj_opt_sim = x_u_traj_opt_sim[:, N*nx:None]

plt.figure(2)
plt.subplot(2,1,1)
plt.plot(x_traj_opt_sim.T)
plt.title('state trajectory')
plt.legend([r'$\theta$', r'$\omega$'])

plt.subplot(2,1,2)
plt.step(range(N), u_traj_opt_sim.T, where='post')
plt.title('control trajectory')
plt.legend([r'$\tau$'])
plt.xlabel('discrete time $k$')


plt.show()
